﻿using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Audio;
using UnityEngine.SceneManagement;

public class Boat : MonoBehaviour {

    public float SteerFactor { get { return steerFactor; } }

    [Header("Steering")]
    [SerializeField] private float steerAccelerationSpeed;
    [SerializeField] private float steerDeccelerationSpeed;
    [SerializeField] private float steerSpeedMax;
    [SerializeField] private float steerRotationMax;

    [Header("Sailing")]
    [SerializeField] private float speedMin;
    [SerializeField] private float speedMax;

    [Header("Balancing")]
    [SerializeField] private float breakerDegreesSafeZone;
    [SerializeField] private int capsizeDrownCountMin;
    [SerializeField] private int capsizeDrownCountMax;
    [SerializeField] private float singleDrownChance;

    [Header("References")]
    [SerializeField] private Transform steerTransform;
    [SerializeField] private Boom boom;
    [SerializeField] private Rope rope;
    [SerializeField] private Capsizer capsizer;
    [SerializeField] private TriggerDetector breakerDetector;
    [SerializeField] private TriggerDetector endingDetector;

    [Header("Audio")]
    [SerializeField] private AudioSource capsizeAudio;
    [SerializeField] private AudioSource reelAudio1;
    [SerializeField] private AudioSource reelAudio2;
    [SerializeField] private float reelAudioIntervalMin;
    [SerializeField] private AudioMixerSnapshot silence;
    [SerializeField] private AudioSource waveAudio;
    [SerializeField] private AudioSource screamAudio;

    private float steerFactor;
    private bool steeredLastFrame;
    private bool hasEnded;

    private float steerRotation;

    private SoundEmitter soundEmitter;
    private float timeSinceLastRopeAudio;
    private bool reelAudioPlayed;
    private float prevAltitude;

	public void Steer(float steer) {
        if (hasEnded) { return; }
        steerFactor += steer * Time.deltaTime * steerAccelerationSpeed;
        steerFactor = Mathf.Clamp(steerFactor, -1.0f, 1.0f);
        if (Mathf.Abs(steer) > .1f) {
            steeredLastFrame = true;
        }
    }

    public void PullRope() {
        if (hasEnded) { return; }
        rope.Pull();
        if (timeSinceLastRopeAudio > reelAudioIntervalMin) {
            timeSinceLastRopeAudio = 0.0f;
            reelAudioPlayed = !reelAudioPlayed;
            if (reelAudioPlayed) {
                soundEmitter.PlayOnce(reelAudio1);
            } else {
                soundEmitter.PlayOnce(reelAudio2);
            }
        }
    }

    public void Capsize(bool toRight) {
        if (hasEnded) { return; }

        GetComponent<SoundEmitter>().PlayOnce(screamAudio);

        CoroutineHelper.Delay(() => {
            int drownCount = UnityEngine.Random.Range(capsizeDrownCountMin, capsizeDrownCountMax);
            for (int i = 0; i < drownCount; i++) {
                DrownAGuy();
            }
        }, 1.0f);

        capsizer.Capsize(toRight, delegate {
            List<Sailor> sailors = new List<Sailor>(FindObjectsOfType<Sailor>());
            sailors.RemoveAll(x => x.HasDrowned);
            sailors.ForEach(x => x.Repriorities());
        });

        soundEmitter.PlayOnce(capsizeAudio);
    }

    public void DrownAGuy() {
        List<Sailor> sailors = new List<Sailor>(FindObjectsOfType<Sailor>());
        sailors.RemoveAll(x => x.HasDrowned);
        if (sailors.Count == 1) { return; }
        sailors.Sort(delegate (Sailor a, Sailor b) { return b.CurrentWorkstation.Priority.CompareTo(a.CurrentWorkstation.Priority); });
        Sailor unluckySailor = sailors[0];
        if (unluckySailor != null) {
            unluckySailor.Drown();
        }
        sailors.Remove(unluckySailor);
    }

    private void Awake() {
        breakerDetector.OnTrigger += OnBreakerTrigger;
        endingDetector.OnTrigger += OnEndingTrigger;

        soundEmitter = GetComponent<SoundEmitter>();
        silence.TransitionTo(0.5f);
    }

    private void Update() {
        timeSinceLastRopeAudio += Time.deltaTime;

        UpdateSteering();
        UpdateMovement();

        if (Input.GetKeyDown(KeyCode.Escape)) {
            RestartGame();
        }
    }

    private void UpdateMovement() {
        if (hasEnded) { return; }
        float speed = Mathf.Lerp(speedMin, speedMax, Mathf.Abs(boom.Power));
        if (Input.GetKey(KeyCode.LeftControl)) {
            speed = speedMax * 5;
        }
        speed *= capsizer.IsCapsized ? .5f : 1.0f;
        transform.Translate(steerTransform.forward.Flattened().normalized * speed * Time.deltaTime);
    }

    private void UpdateSteering() {
        if (!steeredLastFrame) {
            steerFactor = Mathf.MoveTowards(steerFactor, 0.0f, Time.deltaTime * steerDeccelerationSpeed);
        }
        steeredLastFrame = false;
        steerRotation += Time.deltaTime * steerFactor * steerSpeedMax;
        steerRotation = Mathf.Clamp(steerRotation, -steerRotationMax, steerRotationMax);
        steerTransform.localRotation = Quaternion.Euler(0.0f, steerRotation, 0.0f);
    }

    private void OnBreakerTrigger(GameObject breakerObject) {
        Wave wave = breakerObject.GetComponentInParent<Wave>();
        if (wave.HasTriggeredBoat) { return; }
        wave.HasTriggeredBoat = true;

        soundEmitter.PlayOnce(waveAudio);

        float degrees = 90.0f * (Vector3.Dot(wave.transform.forward, -steerTransform.forward) + 1.0f);
        if (degrees < breakerDegreesSafeZone) {
            if (UnityEngine.Random.value < singleDrownChance) {
                DrownAGuy();
            }
            return;
        }

        bool isRight = Vector3.Cross(wave.transform.forward, (-steerTransform.forward).Flattened()).y < 0.0f;

        Capsize(isRight);
    }

    private void OnEndingTrigger(GameObject endingObject) {
        hasEnded = true;
        FindObjectOfType<CameraTarget>().enabled = false;
        FindObjectOfType<EndingCameraPath>().Initiate();
        CoroutineHelper.Delay(() => {
            FindObjectOfType<CameraFader>().FadeOut();
            CoroutineHelper.Delay(() => {
                silence.TransitionTo(5.0f);
                CoroutineHelper.Delay(() => {
                    RestartGame();
                }, 8.0f);
            }, 3.0f);
        }, 3.0f);
    }

    private void RestartGame() {
        Scene scene = SceneManager.GetActiveScene();
        SceneManager.LoadScene(scene.name);
    }

}
